import React, { useState, useRef, useEffect, useLayoutEffect, useCallback } from 'react';
import { Handle, Position, NodeProps } from 'reactflow';
import { SchemaReactflowNodeWithFunctions } from '../../../model/reactflow';
import { QueryElementTypes } from '@graphpolaris/shared/lib/querybuilder';
import { SchemaNode } from '../../../model';
import { EntityPill } from '@graphpolaris/shared/lib/components';
import { Tooltip, TooltipContent, TooltipTrigger } from '@graphpolaris/shared/lib/components/tooltip';
import { VisualizationTooltip } from '@graphpolaris/shared/lib/components/VisualizationTooltip';
import { SchemaPopUp, SchemaPopUpProps } from '../SchemaPopUp/SchemaPopUp';

export const SchemaEntityPill = React.memo(({ id, selected, data }: NodeProps<SchemaReactflowNodeWithFunctions>) => {
  const [openPopupLocation, setOpenPopupLocation] = useState<{ x: number; y: number } | null>(null);

  const ref = useRef<HTMLDivElement>(null);
  /**
   * adds drag functionality in order to be able to drag the entityNode to the schema
   * @param event React Mouse drag event
   */
  const onDragStart = (event: React.DragEvent<HTMLDivElement>) => {
    const eventData: SchemaNode = {
      type: QueryElementTypes.Entity,
      name: id,
      attributes: data.attributes.map((attribute) => attribute.attributes).flat(), // TODO: this seems wrong
    };
    event.dataTransfer.setData('application/reactflow', JSON.stringify(eventData));
    event.dataTransfer.effectAllowed = 'move';
  };

  useEffect(() => {
    if (data.tooltipClose === true) {
      setOpenPopupLocation(null);
    }
  }, [data.tooltipClose]);

  return (
    <>
      <div
        className="w-fit h-fit"
        onDragStart={(event) => onDragStart(event)}
        onDragStartCapture={(event) => onDragStart(event)}
        onMouseDownCapture={(event) => {
          if (!event.shiftKey) event.stopPropagation();
        }}
        onClickCapture={(event) => {
          if (openPopupLocation != null) {
            return;
          }

          const div = event.target as HTMLDivElement;
          const rect = div.getBoundingClientRect();

          setOpenPopupLocation({
            x: data.viewport.x - rect.width / 2,
            y: data.viewport.y - rect.height / 2,
          });
        }}
        draggable
        ref={ref}
      >
        {openPopupLocation !== null && (
          <Tooltip key={data.name} open={true} boundaryElement={data.reactFlowRef} showArrow={true}>
            <TooltipTrigger x={data.viewport.x - openPopupLocation.x} y={data.viewport.y - openPopupLocation.y} />
            <TooltipContent side="right">
              <div>
                <VisualizationTooltip name={data.name} colorHeader={'hsl(var(--clr-node))'}>
                  <SchemaPopUp
                    data={data.attributes.reduce(
                      (acc, attr) => {
                        if (attr.name && attr.type) {
                          acc[attr.name] = attr.type;
                        }
                        return acc;
                      },
                      {} as Record<string, string>,
                    )}
                    numberOfElements={data.nodeCount}
                  />
                </VisualizationTooltip>
              </div>
            </TooltipContent>
          </Tooltip>
        )}

        <EntityPill
          draggable
          title={id}
          withHandles="vertical"
          handleUp={
            <Handle
              style={{ pointerEvents: 'none' }}
              id="entityTargetLeft"
              position={Position.Top}
              className={'!rounded-none !bg-transparent !w-full !h-full !border-0 !b-0 !top-0 !bottom-0'}
              type="target"
            ></Handle>
          }
          handleDown={
            <Handle
              style={{ pointerEvents: 'none' }}
              id="entitySourceRight"
              position={Position.Bottom}
              className={'!rounded-none !bg-transparent !w-full !h-full !border-0 !top-0 !bottom-0'}
              type="source"
            ></Handle>
          }
        />
      </div>
    </>
  );
});

SchemaEntityPill.displayName = 'EntityNode';
